###############################################################################
#
#   MarathUp - Dialog facilities
#
###############################################################################

module Marathon

	#----------------------------------------------------------------------------------
	#
	#   A class for building attribute-editing dialogs. Generates html with
	#   initial values taken from an attribute dictionary, and puts the entered
	#   values back into the dictionary afterwards.
	#
	#----------------------------------------------------------------------------------	

	class DialogDefinition
	
		@@DebugHtml = false
	
		#attr :width
		#attr :height
	
		def initialize(title, width, height, attrs, scroll = false)
			@title = title
			@width = width
			@height = height
			@attributes = attrs
			@scroll = scroll
		end
		
		def html(values, mixed)
			#  There seems to be a bug in Sketchup 6.0 (6.4.10) such that action
			#  callbacks don't work if the html is loaded using set_html() instead
			#  of set_file(). So the best we can do is have the OK button close
			#  the window and trigger the on_close handler, and we can't implement
			#  a Cancel button easily.
			head = <<-END
				<head>
					<style type="text/css">
						* {	font-size: small; font-family: Arial, sans-serif; }
					</style>
					<script language="javascript">
						function turnoff(id) {
							elem = document.getElementById(id);
							elem.checked = false;
						}
						function on_ok() {
							window.close();
						}
					</script>
				</head>
			END
			table_head = '<table>'
			table_rows = @attributes.map {|x|
				format("<tr>%s</tr>", x.html(values, nil, mixed))
			}
			table_foot = '</table>'
			button = '<input type=button value="Accept" onclick="on_ok()">'
			body = '<body>' + table_head + table_rows.join("\n") + table_foot + button + '</body>'
			result = '<html>' + head + body + '</html>'
			if @@DebugHtml
				printf("HTML:\n%s\n", result)
			end
			return result
		end
		
		def get_values(dlog, attrs, mixed)
			for attr in @attributes
				attr.get_values(dlog, attrs, nil, mixed)
			end
		end
		
		def edit(attrs, mixed = false, &proc)
			dlog = UI::WebDialog.new(@title, @scroll)
			html = self.html(attrs, mixed)
			dlog.set_html(html)
			dlog.set_size(@width, @height)
			dlog.set_on_close() {
				self.get_values(dlog, attrs, mixed)
				#dlog.close()
				if proc
					proc.call()
				end
			}
			dlog.show_modal()
		end
		
		def edit_entities(entities, key, named = false, &proc)
			mixed = (entities.length > 1)
			if mixed
				attrs = Marathon.get_multiple_attributes(entities, key)
			else
				attrs = Marathon.get_attributes(entities[0], key, true)
				if named
					attrs["name"] = entities[0].name
				end
			end
			self.edit(attrs, mixed) {
				if mixed
					Marathon.set_multiple_attributes(entities, key, attrs)
				else
					if named
						entities[0].name = attrs["name"]
					end
				end
				if proc
					proc.call()
				end
			}
		end
	
	end # class DialogDefinition
	
	#----------------------------------------------------------------------------------
	#
	#   Attribute classes for use in a DialogDefinition
	#
	#----------------------------------------------------------------------------------
	
	class DialogDecoration
	
		def initialize(html)
			@contents = format('<td colspan="2">%s</td>', html)
		end
		
		def html(attrs, prefix, mixed)
			return @contents
		end
		
		def get_values(dlog, values, prefix, mixed)
		end
	
	end # class DialogDecoration
	
	HRule = DialogDecoration.new('<hr>')

	#----------------------------------------------------------------------------------
	
	class DialogAttr
	
		attr :title
		attr :show_title, true
	
		def initialize(name, title = "")
			@name = name
			@title = title
			@show_title = true
		end
		
		def element_id(prefix)
			id = @name
			if prefix
				id = format("%s.%s", prefix, id)
			end
			return id
		end
		
		def put_value(attrs, id, value, mixed)
			if not mixed and value == nil
				#printf("Deleting %s\n", id.inspect) ###
				attrs.delete_key(id)
			else
				#printf("Putting %s => %s\n", id.inspect, value.inspect) ###
				attrs[id] = value
			end
		end

		def checkbox_html(id, title, current, mixed)
			#printf("Creating checkbox html for title %s current %s mixed %s\n",
			#	title.inspect, current.inspect, mixed.inspect) ###
			if mixed
				script1 = format(" onclick=\"turnoff('%s')\"", id + ".mixed")
				script2 = format(" onclick=\"turnoff('%s')\"", id)
			else
				script1 = ""
			end
			html = format('<input id="%s" type="checkbox" %s %s> %s',
				id, current == true ? "checked" : "", script1, title)
			if mixed
				html = format('<input id="%s.mixed" type="radio" %s%s>%s',
					id, current == :Mixed ? "checked" : "", script2, html)
			end
			return html
		end
	
		def checkbox_value(dlog, id, mixed)
			value = nil
			if mixed
				text = dlog.get_element_value(id + ".mixed")
				if text == "on"
					value = :Mixed
				end
			end
			if value != :Mixed
				text = dlog.get_element_value(id)
				value = (text == "on")
			end
			return value
		end

	end # class DialogAttr

	#----------------------------------------------------------------------------------
	
	class LabelledAttr < DialogAttr
	
		def html(attrs, prefix, mixed)
			return format('<td>%s</td><td>%s</td>',
				@title, self.control_html(attrs, prefix, mixed))
		end
	
	end # class LabelledAttr
	
	#----------------------------------------------------------------------------------
	
	class UnlabelledAttr < DialogAttr

		def html(attrs, prefix, mixed)
			return format('<td colspan="2">%s</td>', self.control_html(attrs, prefix, mixed))
		end
	
	end # class UnlabelledAttr
	
	#----------------------------------------------------------------------------------
	
	class DialogLabel < UnlabelledAttr
	
		def initialize(title)
			@title = title
		end
	
		def control_html(attrs, prefix, mixed)
			return @title
		end
		
		def get_values(dlog, values, prefix, mixed)
		end
	
	end

	#----------------------------------------------------------------------------------
	
	class StrAttr < LabelledAttr
	
		#
		#   String attribute
		#
	
		def initialize(name, title, size = 20)
			super(name, title)
			@size = size
		end

		def control_html(attrs, prefix, mixed)
			id = self.element_id(prefix)
			value =  attrs[id]
			if value == :Mixed
				value = "-"
			end
			if mixed and @name == "name"
				readonly = " readonly"
			else
				readonly = ""
			end
			return format('<input id="%s" type=text size ="%s" value = "%s"%s>',
				id, @size, value, readonly)
		end
		
		def get_values(dlog, values, prefix, mixed)
			id = self.element_id(prefix)
			text = dlog.get_element_value(id).strip()
			#printf("StrAttr.get_values: %s = %s\n", id, text) ###
			if mixed and text == "-"
				value = :Mixed
			else
				value = text
			end
			put_value(values, id, value, mixed)
		end	

	end # class StrAttr

	#----------------------------------------------------------------------------------

	class IntAttr < LabelledAttr
	
		#
		#   Integer attribute
		#
	
		def control_html(attrs, prefix, mixed)
			id = self.element_id(prefix)
			value = attrs[id]
			if value == :Mixed
				value = "-"
			end
			return format('<input id="%s" type=text size ="6" value = "%s">',
				id, value)
		end
		
		def get_values(dlog, values, prefix, mixed)
			id = self.element_id(prefix)
			text = dlog.get_element_value(id).strip()
			#printf("IntAttr.get_values: %s = %s\n", id, text) ###
			if mixed and text == "-"
				value = :Mixed
			elsif text != ""
				value = text.to_i()
			else
				value = nil
			end
			put_value(values, id, value, mixed)
		end
		
	end # class IntAttr

	#----------------------------------------------------------------------------------

	class FloatAttr < LabelledAttr
	
		#
		#   Floating-point number attribute
		#
		
		def control_html(attrs, prefix, mixed)
			id = self.element_id(prefix)
			value = attrs[id]
			if value == :Mixed
				value = "-"
			end
			return format('<input id="%s" type=text size ="6" value = "%s">',
				id, value)
		end
		
		def get_values(dlog, values, prefix, mixed)
			id = self.element_id(prefix)
			text = dlog.get_element_value(id).strip()
			#printf("IntAttr.get_values: %s = %s\n", id, text) ###
			if mixed and text == "-"
				value = :Mixed
			elsif text != ""
				value = text.to_f()
			else
				value = nil
			end
			put_value(values, id, value, mixed)
		end
		
	end # class FloatAttr

	#----------------------------------------------------------------------------------

	class EnumAttr < LabelledAttr
	
		#
		#   Enumerated type attribute, sequentially numbered from 0
		#   Edited as a drop-down menu
		#
	
		def initialize(name, title, values, has_null_value = false)
			super(name, title)
			@values = values
			if has_null_value
				@base_value = 1
			else
				@base_value = 0
			end
		end
		
		def values
			return @values
		end
		
		def control_html(attrs, prefix, mixed)
			id = self.element_id(prefix)
			current = attrs[id]
			#printf("EnumAttr.html: %s: current = %s\n", id, current.inspect) ###
			header = format('<select id="%s">', id)
			values = self.values
			#printf("Building an EnumAttr control with values %s\n", values.inspect) ###
			items = []
			if mixed
				items.push(self.option_html("-", "-", current == :Mixed))
			end
			for i in 0...values.length
				value = i - @base_value
				items.push(self.option_html(value, values[i], value == current))
			end
			footer = '</select>'
			return header + items.join("\n") + footer
		end
		
		def option_html(value, text, selected)
			return format('<option value="%s" %s>%s</option>',
				value, selected ? "selected" : "", text)
		end
		
		def get_values(dlog, attrs, prefix, mixed)
			id = self.element_id(prefix)
			text = dlog.get_element_value(id)
			if mixed and text == "-"
				value = :Mixed
			else
				value = text.to_i()
			end
			#printf("EnumAttr.get_values: %s = '%s' index %s\n", id, text.inspect, value.inspect) ###
			put_value(attrs, id, value, mixed)
		end
	
	end # class EnumAttr

	#----------------------------------------------------------------------------------
	
	class BoolAttr < UnlabelledAttr
	
		#
		#   Boolean attribute, edited as a check box
		#
		
		def control_html(attrs, prefix, mixed)
			id = self.element_id(prefix)
			current = attrs[id]
			if @show_title
				title = " " + @title
			else
				title = ""
			end
			return self.checkbox_html(id, title, current, mixed)
		end
		
		def get_values(dlog, attrs, prefix, mixed)
			id = self.element_id(prefix)
			value = self.checkbox_value(dlog, id, mixed)
			put_value(attrs, id, value, mixed)
		end

	end # class BoolAttr

	#----------------------------------------------------------------------------------
	
	class FlagsAttr < DialogAttr
	
		#
		#   Flags attribute, edited as a set of check boxes
		#
	
		def initialize(name, ncols, values)
			@name = name
			@values = values
			@ncols = ncols
		end
		
		def html(attrs, prefix, mixed)	
			id = self.element_id(prefix)
			current = (attrs[id] or 0)
			if mixed
				current, mask = current
			else
				mask = 0
			end
			rows = []
			for bit, text in @values
				bit_id = self.bit_id(id, bit)
				if (mask & bit) != 0
					bit_on = :Mixed
				else
					bit_on = (current & bit) != 0
				end
				control = self.checkbox_html(bit_id, text, bit_on, mixed)
				rows.push(control)
			end
			columns = []
			n = rows.length
			nrows = (n + @ncols - 1) / @ncols
			i = 0
			while i < n
				controls = rows[i, nrows]
				columns.push('<td>' + controls.join("<br>\n") + '</td>')
				i += nrows
			end
			sep = '<td width="20"></td>'
			table = '<table cellspacing="0" cellpadding="0"><tr>' + columns.join(sep) + '</tr></table>'
			return format('<td colspan="2">' + table + '</td>')
		end
		
		def get_values(dlog, attrs, prefix, mixed)
			id = self.element_id(prefix)
			value = 0
			mask = 0
			for bit, text in @values
				bit_id = self.bit_id(id, bit)
				bit_on = self.checkbox_value(dlog, bit_id, mixed)
				if bit_on == :Mixed
					mask |= bit
				elsif bit_on
					value |= bit
				end
			end
			#printf("FlagsAttr.get_values: new value = 0x%x\n", value.inspect) ###
			if mixed
				value = [value, mask]
			end
			put_value(attrs, id, value, mixed)
		end
		
		def bit_id(id, bit)
			return format("%s:%s", id, bit)
		end
	
	end # class FlagsAttr

end # module Marathon
