##################################################################################
#
#   Sketchup to Marathon Export Script - Data Structures
#
##################################################################################

class MarathonExporter

	#---------------------------------------------------------------------------------
	
	class ListMap
	
		def initialize()
			@map = {}
		end
		
		def [](key)
			if @map.has_key?(key)
				entry = @map[key]
			else
				entry = []
				@map[key] = entry
			end
			return entry
		end
		
		def each
			for x in @map
				yield x
			end
		end
	
	end # class ListMap
	
	#---------------------------------------------------------------------------------
	
	class WallDescriptor
	
		attr :face, true            #  Face
		attr :ctx, true           #  Context
		attr :point_numbers, true   #  [i, j] with i < j
		attr :heights               #  [z0, z1]
		attr :sides                 #  [WallSide, WallSide]
		
		def initialize()
			@heights = [0, 0]
			@sides = [nil, nil]
		end
		
		def edges()
			e = @point_numbers
			return [[e, @heights[BOTTOM]], [e, @heights[TOP]]]
		end
		
		def inspect()
			return format("<WallDesc %s bh=%s th=%s rs=%s ls=%s>",
				@point_numbers.inspect,
				@heights[BOTTOM].inspect,
				@heights[TOP].inspect,
				@sides[RIGHT].inspect,
				@sides[LEFT].inspect)
		end
	
	end # class WallDescriptor
	
	#---------------------------------------------------------------------------------
	
	class WallSide
	
		attr :material, true
		attr :uvs
		attr :face, true
		attr :ctx, true
		
		def initialize()
			@uvs = [nil, nil]
		end
		
		def inspect()
			mat = @material
			if mat != nil
				matname = mat.name
			else
				matname = nil
			end
			return format("<WallSide mat=%s buv=%s tuv=%s>",
				matname.inspect,
				@uvs[BOTTOM].inspect,
				@uvs[TOP].inspect)
		end
	
	end # class WallSide
	
	#---------------------------------------------------------------------------------
	
	class MarathonLine
	
		attr :number
		attr :point_numbers        #  [i, j] with i < j
		attr :polygons             #  [MarathonPoly, MarathonPoly]
		attr :sides                #  [MarathonSide, MarathonSide]
		attr :upper_wall, true     #  WallDescriptor
		attr :lower_wall, true     #  WallDescriptor
		attr :spanning_wall, true  #  WallDescriptor
		attr :flags, true
		
		def initialize(n, points)
			@number = n
			@point_numbers = points
			@walls = []
			@polygons = [nil, nil]
			@sides = [nil, nil]
		end
		
		def upper_wall_side(which_side)
			return wall_side(self.upper_wall, which_side)
		end
		
		def lower_wall_side(which_side)
			return wall_side(self.lower_wall, which_side)
		end
		
		def spanning_wall_side(which_side)
			return wall_side(self.spanning_wall, which_side)
		end
		
		def wall_side(wall, which_side)
			if wall
				return wall.sides[which_side]
			else
				return nil
			end
		end
		
		def attributes
			wall = (@spanning_wall or @lower_wall or @upper_wall)
			if wall
				return Marathon.get_wall_attributes(wall.face)
			else
				return {}
			end
		end
		
		def merged_wall_line_flags()
			flags = 0
			for wall in [@upper_wall, @lower_wall, @spanning_wall]
				if wall
					attrs = Marathon.get_wall_attributes(wall.face)
					flags |= (attrs["line_flags"] or 0)
				end
			end
			return flags
		end
		
		def connect_to_poly(poly, edge_no)
			i, j = poly.edge_point_numbers(edge_no)
			if i < j
				@polygons[CLOCKWISE] = poly
			else
				@polygons[ANTICLOCKWISE] = poly
			end
			poly.lines[edge_no] = self
		end
		
		def length(points)
			x0, y0 = points[@point_numbers[0]]
			x1, y1 = points[@point_numbers[1]]
			dx = x1 - x0
			dy = y1 - y0
			h = Math.sqrt(dx * dx + dy * dy)
			return h.round
		end
		
		def highest_adjacent_floor()
			h = self.adjacent_heights(FLOOR)
			if h
				return h.max()
			else
				return 0
			end
		end
		
		def lowest_adjacent_ceiling()
			h = self.adjacent_heights(CEILING)
			if h
				return h.min()
			else
				return 0
			end
		end
		
		def adjacent_heights(which)
			h = []
			for poly in @polygons
				if poly
					h.push(poly.heights[which])
				end
			end
			return h
		end
		
		def other_poly(poly)
			i = @polygons.index(poly)
			return @polygons[1 - i]
		end
		
		def side_for_poly(poly)
			i = @polygons.index(poly)
			return @sides[i]
		end
		
		def side_for_other_poly(poly)
			i = @polygons.index(poly)
			return @sides[1 - i]
		end
		
		def representative_wall
			return (@spanning_wall or @upper_wall or @lower_wall)
		end
		
		def inspect()
			return format("<MarathonLine %d>", @number)
		end
		
		def dump()
			return format("<MarathonLine %d %s cwp=%s acwp=%s cws=%s acws=%s uw=%s sw=%s lw=%s>",
				@number,
				@point_numbers.inspect,
				@polygons[CLOCKWISE].inspect,
				@polygons[ANTICLOCKWISE].inspect,
				@sides[CLOCKWISE].inspect,
				@sides[ANTICLOCKWISE].inspect,
				@upper_wall.inspect,
				@spanning_wall.inspect,
				@lower_wall.inspect)
		end
	
	end # class MarathonLine
	
	#---------------------------------------------------------------------------------
	
	class MarathonSideTextureDefinition
	
		NullBytes = [0, 0, -1, -1].pack("nnCC")
		
		attr :x0, true
		attr :y0, true
		attr :texture
		attr :transfer_mode
		attr :light
	
		def initialize(x0, y0, texture, mode, light)
			@x0 = x0
			@y0 = y0
			@texture = texture
			@transfer_mode = mode
			@light = light
		end
		
		def MarathonSideTextureDefinition.from_material(uv, material, mode, light)
			shape = MarathonShapeDescriptor.from_name(material.name)
			if shape.is_landscape
				mode = 9 # landscape
			end
			return MarathonSideTextureDefinition.new(
				marathon_coord(uv.x) % 1024,
				-marathon_coord(uv.y) % 1024,
				shape,
				mode,
				light)
		end
		
		def next_texture()
			return MarathonSideTextureDefinition.new(
				@x0, @y0, @texture.next_texture(), @transfer_mode, @light)
		end
		
		def inspect()
			return format("<Texture %s %s %s>",
				@x0, @y0, @texture.inspect)
		end
	
	end # class MarathonSideTextureDefinition
	
	#---------------------------------------------------------------------------------
	
	class MarathonShapeDescriptor
	
		NullBytes = [-1, -1].pack("CC")
		
		attr :set
		attr :frame
		attr :is_landscape
	
		def initialize(set, frame, is_landscape = false)
			@set = set
			@frame = frame
			@is_landscape = is_landscape
		end

		def MarathonShapeDescriptor.from_name(name)
			set, frame, suffix = name.split("-")
			MarathonShapeDescriptor.new(set.to_i, frame.to_i, suffix == "L")
		end
		
		def next_texture()
			return MarathonShapeDescriptor.new(@set, @frame + 1)
		end
		
		def inspect()
			return format("<Shape set=%s frame=%s>",
				@set, @frame)
		end
	
	end # class MarathonShapeDescriptor
	
	#---------------------------------------------------------------------------------
	
	NoShape = MarathonShapeDescriptor.new(-1, -1)
	NoTexture = MarathonSideTextureDefinition.new(0, 0, NoShape, 0, nil)
			
	#---------------------------------------------------------------------------------
	
	FULL_SIDE = 0
	HIGH_SIDE = 1
	LOW_SIDE = 2
	COMPOSITE_SIDE = 3
	SPLIT_SIDE = 4
	
	class MarathonSide
	
		attr :number
		attr :type, true
		attr :primary_texture, true      #  MarathonSideTextureDefinition
		attr :secondary_texture, true    #  MarathonSideTextureDefinition
		attr :transparent_texture, true  #  MarathonSideTextureDefinition
		attr :polygon, true              #  MarathonPoly
		attr :line, true                 #  MarathonLine
		attr :flags, true
		attr :permutation_type, true     # 'platform', 'light', 'tag', 'terminal'
		attr :permutation_name, true     #  evaluated
		attr :attrs, true
		
		def initialize(n)
			@number = n
			@type = 0
		end
		
		def attributes
			return @attrs
		end
		
		def primary_light
			light = self.default_light(@primary_texture)
			if MarathonExporter.DebugLights
				printf("Primary light of side %s is %s\n",
					@number, light.inspect)
			end
			return light
		end
		
		def secondary_light
			return self.default_light(@secondary_texture)
		end
		
		def transparent_light
			return self.default_light(@transparent_texture)
		end
		
		def default_light(texture)
			light = texture.light
			if light == ""
				light = @polygon.default_light
			end
			return light
		end
		
		def adjust_upper_texture_offset(dy)
			case @type
				when FULL_SIDE, HIGH_SIDE, SPLIT_SIDE
					#printf("Adjusting texture offset of %s by %s\n",
					#	self.inspect, dy.inspect) ###
					texture = @primary_texture
					texture.y0 += dy
			end
		end
		
		def inspect()
			return format("<Side %d>", @number)
		end
		
		def dump()
			return format("<Side %s type=%s pt=%s st=%s tt=%s poly=%s line=%s>",
				@number,
				@type,
				@primary_texture.inspect,
				@secondary_texture.inspect,
				@transparent_texture.inspect,
				@polygon.inspect,
				@line.inspect)
		end
	
	end # class MarathonSide
	
	#---------------------------------------------------------------------------------
	
	class MarathonObject
	
		attr :number
		attr :type, true
		attr :index, true
		attr :facing, true
		attr :polygon, true
		attr :location, true
		attr :flags, true
		attr :inst, true   # ComponentInstance
		attr :ctx, true  # Context
		
		def initialize(n)
			@number = n
		end
		
		def inspect()
			return format("<Obj %d>", @number)
		end
		
		def dump()
			return format("<Obj %d type=%s index=%s facing=%s poly=%s loc%s>",
				@number,
				@type,
				@index,
				@facing,
				@polygon.inspect,
				@location.inspect)
		end
	
	end # class MarathonObject
	
	#---------------------------------------------------------------------------------
	
	class MarathonPoly
	
		attr :number
		attr :name, true
		attr :floor_ctx, true
		attr :ceiling_ctx, true
		attr :floor, true
		attr :ceiling, true
		attr :point_numbers           # Clockwise list of point nos.
		attr :lines                   # List of MarathonLine
		attr :floor_height, true
		attr :ceiling_height, true
		attr :floor_texture, true     # MarathonShapeDescriptor
		attr :ceiling_texture, true   # MarathonShapeDescriptor
		attr :floor_origin, true      # [x0, y0]
		attr :ceiling_origin, true    # [x0, y0]
		attr :floor_light, true       # evaluated
		attr :ceiling_light, true     # evaluated
		attr :floor_transfer_mode, true
		attr :ceiling_transfer_mode, true
		attr :objects
		attr :area, true
		attr :center, true
		attr :platform, true          # MarathonPlatform
		attr :permutation_name, true  # evaluated
		attr :media_name, true        # evaluated
		attr :media_light, true       # evaluated
		attr :ambient_sound, true     # evaluated
		attr :random_sound, true      # evaluated
		
		
		def initialize(n, points)
			@number = n
			@point_numbers = points
			np = points.length
			@lines = Array.new(8)
			@objects = [] # List of MarathonObject
		end
		
		def type
			return (@floor.get_attribute("polygon", "type") or 0)
		end
		
		def num_points
			return @point_numbers.length
		end
		
		def point_number(i)
			# Point number of vertex i mod number of points
			return @point_numbers[i % self.num_points]
		end
		
		def edge_point_numbers(i)
			# Point numbers for edge i, in clockwise direction
			j = (i + 1) % self.num_points
			p = @point_numbers
			return [p[i], p[j]]
		end
		
		def heights
			return [@floor_height, @ceiling_height]
		end
		
		def sides()
			result = []
			for line in @lines
				if line
					result.push(line.side_for_poly(self))
				else
					result.push(nil)
				end
			end
			return result
		end
		
		def adjacent_polys()
			result = []
			for line in @lines
				if line
					result.push(line.other_poly(self))
				else
					result.push(nil)
				end
			end
			return result
		end
		
		def default_light
			light = self.ceiling_light
			if MarathonExporter.DebugLights
				printf("Default light of poly %s = %s\n",
					@number, light)
			end
			return light
		end
		
		def ceiling_light
			light = @ceiling_light
			if light == ""
				light = @floor_light
			end
			if MarathonExporter.DebugLights
				printf("Ceiling light of poly %s = %s\n",
					@number, light)
			end
			return light
		end
		
		def floor_light
			light = @floor_light
			if light == ""
				light = @ceiling_light
			end
			if MarathonExporter.DebugLights
				printf("Floor light of poly %s = %s\n",
					@number, light)
			end
			return light
		end
		
		def inspect()
			return format("<MarathonPoly %d>", @number)
		end
		
		def dump()
			return format("<MarathonPoly %d points=%s floor_height=%s ceiling_height=%s lines=%s area=%s center=%s>",
				@number,
				@point_numbers.inspect,
				@floor_height,
				@ceiling_height,
				@lines.inspect,
				@area.inspect,
				@center.inspect)
		end
	
	end # class MarathonPoly
	
	#---------------------------------------------------------------------------------

	COMES_FROM_FLOOR = 0x00000040
	COMES_FROM_CEILING = 0x00000080

	class MarathonPlatform
	
		attr :number
		attr :polygon
		attr :flags, true
		attr :tag_name, true
	
		def initialize(n, poly)
			@number = n
			@polygon = poly
		end
		
		def name
			name = self.attributes["name"]
			if name == ""
				name = nil
			end
			return name
		end
		
		def attributes
			return Marathon.get_platform_attributes(@polygon.floor)
		end
	
	end # class MarathonPlatform

	#---------------------------------------------------------------------------------

	class MarathonLight
	
		attr :attrs, true
		attr :tag_name, true  # evaluated
	
	end # class MarathonLight

	#---------------------------------------------------------------------------------

	class MarathonMedia
	
		attr :inst, true
		attr :ctx, true
		attr :tide_name, true

	end # class MarathonMedia

	#---------------------------------------------------------------------------------

	class Context
	
		attr :parent, true      # Context
		attr :trans, true       # Transformation
		attr :env, true         # {name => value}
		
		def initialize(parent = nil, obj = nil)
			env = {}
			if parent
				@parent = parent
				@trans = parent.trans * obj.transformation
				attrs = obj.attribute_dictionary("parameters")
				if attrs
					for name, value in attrs
						env[name] = parent.eval_param(value)
					end
				end
			else
				@trans = Geom::Transformation.new()
			end
			@env = env
		end
		
		def eval_param(expr)
			if not expr or expr == ""
				return ""
			elsif expr[0, 1] == "$"
				return self.lookup(expr[1, expr.length - 1])
			else
				return expr
			end
		end
		
		def lookup(name)
			value = @env[name]
			if not value
				if parent = @parent
					value = parent.lookup(name)
				else
					raise Marathon::ExportError.new(format("Parameter '%s' not defined", name))
				end
			end
			return value
		end
	
	end # class Context

	#---------------------------------------------------------------------------------

end # class MarathonExporter
